import 'dart:async';
import 'dart:convert';
import 'dart:io';

import 'package:flutter/services.dart';
import 'package:octant/models/user_coords_response.dart';
import 'package:octant/models/user_rank.dart';
import 'package:permission_handler/permission_handler.dart';

class Octant {
  static const String methodChannelInit = 'init';
  static const String methodChannelUpdateUserDetails = 'updateUserDetails';
  static const String methodChannelStartService = 'startService';
  static const String methodChannelStartTracing = 'startTracing';
  static const String methodChannelStopTracing = 'stopTracing';
  static const String methodChannelStopService = 'stopService';
  static const String methodChannelConfigureIntervals = 'configureIntervals';

  static const String methodChannelQueryRanking = 'queryRanking';
  static const String methodChannelQueryUserCoordinate = 'queryUserCoordinate';

  static const String methodChannelEnableDebug = 'enableDebug';

  static const String argumentBaseUrl = 'baseUrl';
  static const String argumentUserCode = 'userCode';
  static const String argumentTeamCode = 'teamCode';

  static const String argumentUserCodes = 'userCodes';
  static const String argumentStart = 'start';
  static const String argumentEnd = 'end';
  static const String argumentLimit = 'limit';
  static const String argumentRemarkCode = 'remarkCode';

  static const String argumentGatheringInterval = 'gatheringInterval';
  static const String argumentUploadingInterval = 'uploadingInterval';
  static const String argumentNotificationIconId = 'notificationIconId';

  static const String argumentNotificationTitle = "notificationTitle";
  static const String argumentNotificationContent = "notificationContent";

  static const String argumentEnableDebug = 'enableDebug';

  static const MethodChannel _channel = MethodChannel('octant');
  static String? userId;
  static String? teamCode;

  ///Setup SDK with userId, teamCode and BaseUrl before we start using SDK
  ///
  ///if userId and TeamCode were not configured by calling init, [startService] method will returns RuntimeError that indicates Octant has not been initialized
  static Future init(String userId, {String? teamCode, String? baseUrl}) async {
    Octant.userId = userId;
    Octant.teamCode = teamCode;
    requestForPermisions();
    await _channel.invokeMethod(methodChannelInit, {
      argumentUserCode: userId,
      argumentTeamCode: teamCode,
      argumentBaseUrl: baseUrl
    });
  }

  ///API to update userId in case we want to switch account
  static Future updateUserDetails(String userId, {String? teamCode}) async {
    Octant.userId = userId;
    Octant.teamCode = teamCode;
    bool? isServiceBound = await _channel.invokeMethod(
        methodChannelUpdateUserDetails,
        {argumentUserCode: userId, argumentTeamCode: teamCode});
    if (isServiceBound ?? false) {
      return Future.value(isServiceBound);
    } else {
      return Future.error('OctantService is not started or registered');
    }
  }

  ///Start a foregroud service to fetch and upload location periodically, the interval can be configured by [configureIntervals]
  ///
  /// For Android version higher than Android.Q, a notification is requried to start a forground service, therefore we are required to provide
  /// [notificationDrawable], [notificationTitle] and [notificationContent] that will be displayed in notification tray while the foreground service is running.
  ///
  /// [notificationDrawable] is the name of drawable, no matter the drawable is R.drawable.ic_launcher or R.mipmap.ic_launcher, the name will be "ic_launcher" only
  static Future<bool?> startService(
      {String? notificationDrawable,
      String? notificationTitle,
      String? notificationContent}) async {
    if (userId != null && teamCode != null) {
      bool? succeed =
          await _channel.invokeMethod<bool>(methodChannelStartService, {
        argumentNotificationIconId: notificationDrawable,
        argumentNotificationTitle: notificationTitle,
        argumentNotificationContent: notificationContent
      }).timeout(const Duration(seconds: 5), onTimeout: () {
        return false;
      });
      return succeed;
    }
    return Future.error(
        'userId and teamCode must not be null, have you called Octant.init(userId, teamCode) before calling startService()');
  }

  ///Start to collecting location and timerTask to upload locations.
  ///
  /// make sure the Serivce is running by calling [startService] before this method
  static Future startTracing() async {
    bool? isServiceBound =
        await _channel.invokeMethod(methodChannelStartTracing);
    if (isServiceBound ?? false) {
      return Future.value(isServiceBound);
    } else {
      return Future.error('OctantService is not started or registered');
    }
  }

  ///Unbind and stop the service
  static Future stopService() async {
    bool? isServiceBound =
        await _channel.invokeMethod(methodChannelStopService);
    if (isServiceBound ?? false) {
      return Future.value(isServiceBound);
    } else {
      return Future.error('OctantService is not started or registered');
    }
  }

  ///Configure the intervals of location gathering and uploading
  ///
  /// [gatheringInterval] frequency of location collection, in second, the minimum interval will be 5s, any value is not 5s' multiple will be rounded.
  /// [uploadingInterval] frequency of location uploading, in second, the minimum of this parameter will be 5 times of [gatheringFrequency]
  static Future configureIntervals(
      {int gatheringInterval = 5, int uploadingInterval = 20}) async {
    bool? isServiceBound =
        await _channel.invokeMethod(methodChannelConfigureIntervals, {
      argumentGatheringInterval: gatheringInterval,
      argumentUploadingInterval: uploadingInterval
    });
    if (isServiceBound ?? false) {
      return Future.value(isServiceBound);
    } else {
      return Future.error('OctantService is not started or registered');
    }
  }

  ///API to query current ranking of travelled distance
  static Future<List<UserRank>> queryRanking(
      {int startTime = 0,
      int endTime = 0,
      int limit = 10,
      List<String> userCodeList = const [],
      String? remarkCode}) async {
    String userCodes = userCodeList.join(',');
    List<dynamic>? ranks =
        await _channel.invokeMethod(methodChannelQueryRanking, {
      argumentStart: startTime.toString(),
      argumentEnd: endTime.toString(),
      argumentLimit: limit.toString(),
      argumentRemarkCode: remarkCode,
      argumentUserCodes: userCodes
    });
    List<UserRank> rankList =
        ranks?.map((e) => UserRank.fromJson(e)).toList() ?? <UserRank>[];

    return rankList;
  }

  ///API to query current user's history locations
  static Future queryUserCoordinate(
      {int startTime = 0,
      int endTime = 0,
      int limit = 10,
      String? remarkCode}) async {
    String? jsonString =
        await _channel.invokeMethod(methodChannelQueryUserCoordinate, {
      argumentStart: startTime.toString(),
      argumentEnd: endTime.toString(),
      argumentLimit: limit.toString(),
      argumentRemarkCode: remarkCode,
    });

    String source = jsonString ?? "";
    Map<String, dynamic>? jsonMap = json.decode(source);
    UserCoordinateResponse userCoordinateResponse =
        UserCoordinateResponse.fromJson(jsonMap);
    return userCoordinateResponse;
  }

  ///Enable debugging logs
  static enableDebug(bool debug) async {
    await _channel
        .invokeMethod(methodChannelEnableDebug, {argumentEnableDebug: debug});
  }

  static Future requestForPermisions() async {
    if (Platform.isAndroid) {
      await Permission.location.request();
      await Permission.locationAlways.request();
    } else if (Platform.isIOS) {}
  }
}
